

#include <stdint.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <MyStruct.h>

#define MAX_NUM_STRUCTS				10
#define FIRST_ASSIGNED_STRUCT_ID 	0xA8F4

static uint16_t NumCreatedStructs = 0;

static volatile MyStruct * volatile * volatile MyStructsPtrs = NULL;

uint16_t MyStruct_Init(void)
{
	if ((MyStructsPtrs = (volatile MyStruct * volatile *) malloc(MAX_NUM_STRUCTS * sizeof(MyStruct *))) == NULL)
		return 1;
	else
	{
		for (uint16_t i = 0; i < MAX_NUM_STRUCTS; i++)
			MyStructsPtrs[i] = NULL;

		return 0;
	}
}

uint16_t MyStruct_Terminate(void)
{
	if (MyStructsPtrs == NULL)
		return 0;

	for (uint16_t i = 0; i < NumCreatedStructs; i++)
	{
		free((void *) MyStructsPtrs[i]);
	}

	free((void *) MyStructsPtrs);

	return 0;
}

uint16_t MyStruct_Create(uint16_t * StructIDptr)
{
	volatile MyStruct * volatile MyStructPtr;

	if (NumCreatedStructs >= MAX_NUM_STRUCTS)
		return 1;

	if ((MyStructPtr = (volatile MyStruct *) malloc(sizeof(MyStruct))) == NULL)
		return 1;

	MyStructsPtrs[NumCreatedStructs] = MyStructPtr;

	*StructIDptr = FIRST_ASSIGNED_STRUCT_ID + NumCreatedStructs;

	// other struct initialization stuff here

	return 0;
}

uint16_t MyStruct_SetPointers(uint16_t MyStructID, size_t FirstPointerOffset, ...)
{
#if defined(DEBUG)
	if ((MyStructID - FIRST_ASSIGNED_STRUCT_ID + 1) > NumCreatedStructs)
			return 1;
#endif

	// Retrieve the address of the MyStruct structure associated to the passed identifier.
	volatile MyStruct * volatile StructPtr = MyStructsPtrs[MyStructID - FIRST_ASSIGNED_STRUCT_ID];

	// Calculate the address of the 'first_pointer' member of StructPtr.
	void * volatile * volatile MyStructPointersArray = (void * volatile *) (((uintptr_t) StructPtr) + FirstPointerOffset);

	va_list ExtraArgs;
	void * volatile CurrentPtr;

	// Initialize the list of variable arguments.
	va_start(ExtraArgs, FirstPointerOffset);

	for (uint16_t i = 0; (CurrentPtr = va_arg(ExtraArgs, void *)) != ((void *) UINTPTR_MAX) && (FirstPointerOffset + i * sizeof(void *)) < sizeof(MyStruct); i++)
	{
		MyStructPointersArray[i] = CurrentPtr;
	}

	// Terminate the list of callback pointers.
	va_end(ExtraArgs);

	return (CurrentPtr != ((void *) UINTPTR_MAX));
}
